﻿using System;
using System.Collections.Generic;
using System.Text;

using NATS.Client;

namespace TestCommon
{
    /// <summary>
    /// 메시지 클라이언트
    /// </summary>
    public class MessageClient : IDisposable
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 서버 발견시 - ServerDiscovered

        /// <summary>
        /// 서버 발견시
        /// </summary>
        public event EventHandler<ServerDiscoveredEventArgs> ServerDiscovered;

        #endregion
        #region 서버 재연결시 - ServerReconnected

        /// <summary>
        /// 서버 재연길시
        /// </summary>
        public event EventHandler<ServerReconnectedEventArgs> ServerReconnected;

        #endregion
        #region 메시지 수신시 - MessageReceived

        /// <summary>
        /// 메시지 수신시
        /// </summary>
        public event EventHandler<MessageReceivedEventArgs> MessageReceived;

        #endregion
        #region 서버 연결 중단시 - ServerDisconnected

        /// <summary>
        /// 서버 연결 중단시
        /// </summary>
        public event EventHandler<ServerDisconnectedEventArgs> ServerDisconnected;

        #endregion
        #region 연결 폐쇄시 중단시 - ConnectionClosed

        /// <summary>
        /// 연결 폐쇄시 중단시
        /// </summary>
        public event EventHandler<ConnectionClosedEventArgs> ConnectionClosed;

        #endregion
        #region 에러시 - Error

        /// <summary>
        /// 에러시
        /// </summary>
        public event EventHandler<ErrorEventArgs> Error;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 서버 정보
        /// </summary>
        private ServerInfo serverInfo = null;

        /// <summary>
        /// 구독 정보 딕셔너리
        /// </summary>
        private Dictionary<string, SubscriptionInfo> subscriptionInfoDictionary = null;

        /// <summary>
        /// 서버 발견시 이벤트 핸들러
        /// </summary>
        private EventHandler<ConnEventArgs> serverDiscoveredEventHandler = null;

        /// <summary>
        /// 서버 재연결시 이벤트 핸들러
        /// </summary>
        private EventHandler<ConnEventArgs> serverReconnectedEventHandler = null;

        /// <summary>
        /// 메시지 수신시 이벤트 핸들러
        /// </summary>
        private EventHandler<MsgHandlerEventArgs> messageReceivedEventHandler = null;

        /// <summary>
        /// 서버 연결 중단시 이벤트 핸들러
        /// </summary>
        private EventHandler<ConnEventArgs> serverDisconnectedEventHandler = null;

        /// <summary>
        /// 연결 폐쇄시 이벤트 핸들러
        /// </summary>
        private EventHandler<ConnEventArgs> connectionClosedEventHandler = null;

        /// <summary>
        /// 에러시 이벤트 핸들러
        /// </summary>
        private EventHandler<ErrEventArgs> errorEventHandler = null;

        /// <summary>
        /// 작동 여부
        /// </summary>
        private bool isWorking = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MessageClient()

        /// <summary>
        /// 생성자
        /// </summary>
        public MessageClient()
        {
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 초기화 하기 - Initialize(serverInfo, subscriptionInfoList)

        /// <summary>
        /// 초기화 하기
        /// </summary>
        /// <param name="serverInfo">서버 정보</param>
        /// <param name="subscriptionInfoList">구독 정보 리스트</param>
        public void Initialize(ServerInfo serverInfo, IList<SubscriptionInfo> subscriptionInfoList)
        {
            this.serverInfo = serverInfo;

            #region 구독 정보 딕셔너리를 설정한다.

            this.subscriptionInfoDictionary = new Dictionary<string, SubscriptionInfo>();

            if(subscriptionInfoList != null)
            {
                foreach(SubscriptionInfo info in subscriptionInfoList)
                {
                    if(this.subscriptionInfoDictionary.ContainsKey(info.Key))
                    {
                        continue;
                    }
                    else
                    {
                        this.subscriptionInfoDictionary.Add(info.Key, info);
                    }
                }
            }

            #endregion

            #region 서버 발견시 이벤트 핸들러를 설정한다.

            this.serverDiscoveredEventHandler = (sender, e) =>
            {
                FireServerDiscoveredEvent(e.Conn);
            };

            #endregion
            #region 서버 재연결시 이벤트 핸들러를 설정한다.

            this.serverReconnectedEventHandler = (sender, e) =>
            {
                FireServerReconnectedEvent(e.Conn);
            };

            #endregion
            #region 메시지 수신시 이벤트 핸들러를 설정한다.

            this.messageReceivedEventHandler = (sender, e) =>
            {
                FireMessageReceivedEvent(e.Message);
            };

            #endregion
            #region 서버 연결 중단시 이벤트 핸들러를 설정한다.

            this.serverDisconnectedEventHandler = (sender, e) =>
            {
                FireServerDisconnectedEvent(e.Conn);
            };

            #endregion
            #region 연결 폐쇄시 이벤트 핸들러를 설정한다.

            this.connectionClosedEventHandler = (sender, e) =>
            {
                FireConnectionClosedEvent(e.Conn);
            };

            #endregion
            #region 에러 이벤트 핸들러를 설정한다.

            this.errorEventHandler = (sender, e) =>
            {
                FireErrorEvent(e);
            };

            #endregion
        }

        #endregion

        #region 시작하기 - Start()

        /// <summary>
        /// 시작하기
        /// </summary>
        public void Start()
        {
            if(this.isWorking)
            {
                return;
            }

            this.isWorking = true;

            try
            {
                Options options = this.serverInfo.GetOptions();

                options.ServerDiscoveredEventHandler = this.serverDiscoveredEventHandler;
                options.ReconnectedEventHandler      = this.serverReconnectedEventHandler;
                options.DisconnectedEventHandler     = this.serverDisconnectedEventHandler;
                options.ClosedEventHandler           = this.connectionClosedEventHandler;
                options.AsyncErrorEventHandler       = this.errorEventHandler;

                this.serverInfo.StartConnection(options);

                foreach(KeyValuePair<string, SubscriptionInfo> keyValuePair in this.subscriptionInfoDictionary)
                {
                    SubscriptionInfo info = keyValuePair.Value;

                    info.StartSubscription(this.serverInfo.Connection, this.messageReceivedEventHandler);
                }
            }
            catch(Exception exception)
            {
                this.isWorking = false;

                throw exception;
            }
        }

        #endregion
        #region 중단하기 - Stop()

        /// <summary>
        /// 중단하기
        /// </summary>
        public void Stop()
        {
            if(!this.isWorking)
            {
                return;
            }

            this.isWorking = false;

            foreach(KeyValuePair<string, SubscriptionInfo> keyValuePair in this.subscriptionInfoDictionary)
            {
                SubscriptionInfo info = keyValuePair.Value;

                if(info != null)
                {
                    info.StopSubscription();
                }
            }

            this.serverInfo.StopConnection();

            Options options = ConnectionFactory.GetDefaultOptions();

            options.ServerDiscoveredEventHandler = null;
            options.ReconnectedEventHandler      = null;
            options.DisconnectedEventHandler     = null;
            options.ClosedEventHandler           = null;
            options.AsyncErrorEventHandler       = null;
        }

        #endregion

        #region 메시지 송신하기 - SendMessage(subject, messageByteArray)

        /// <summary>
        /// 메시지 송신하기
        /// </summary>
        /// <param name="subject">수신 서브젝트</param>
        /// <param name="messageByteArray">송신 메시지 바이트 배열</param>
        public void SendMessage(string subject, byte[] messageByteArray)
        {
            if(string.IsNullOrWhiteSpace(subject))
            {
                throw new ArgumentException("subject가 null이거나 공백 문자열 입니다.");
            }

            if(messageByteArray == null || messageByteArray.Length == 0)
            {
                throw new ArgumentException("messageByteArray가 null이거나 길이가 0 입니다.");
            }

            int tryCount = 0;

            do
            {
                try
                {
                    this.serverInfo.Connection.Publish(subject, messageByteArray);

                    this.serverInfo.Connection.Flush();
                }
                catch(Exception exception)
                {
                    if(tryCount >= 3)
                    {
                        throw exception;
                    }
                }

                tryCount++;
            }
            while(true);
        }

        #endregion
        #region 메시지 송신하기 - SendMessage(subject, message)

        /// <summary>
        /// 메시지 송신하기
        /// </summary>
        /// <param name="subject">수신 서브젝트</param>
        /// <param name="message">송신 메시지</param>
        public void SendMessage(string subject, string message)
        {
            byte[] messageByteArray = Encoding.UTF8.GetBytes(message);
            
            SendMessage(subject, messageByteArray);
        }

        #endregion

        #region 메시지 요청하기 - RequestMessage(subject, messageByteArray, timeout)

        /// <summary>
        /// 메시지 요청하기
        /// </summary>
        /// <param name="subject">수신 서브젝트</param>
        /// <param name="messageByteArray">송신 메시지 바이트 배열</param>
        /// <param name="timeout">타임아웃 (디폴트 : 30초)</param>
        /// <returns>응답 메시지 바이트 배열</returns>
        public byte[] RequestMessage(string subject, byte[] messageByteArray, int timeout = 30000)
        {
            if(string.IsNullOrWhiteSpace(subject))
            {
                throw new ArgumentException("subject가 null이거나 공백 문자열 입니다.");
            }

            if(messageByteArray == null || messageByteArray.Length == 0)
            {
                throw new ArgumentException("messageByteArray가 null이거나 길이가 0 입니다.");
            }

            int tryCount = 0;

            do
            {
                try
                {
                    Msg replyMessage = this.serverInfo.Connection.Request(subject, messageByteArray, timeout);

                    return replyMessage.Data;
                }
                catch(Exception exception)
                {
                    if(tryCount >= 3)
                    {
                        throw exception;
                    }
                }

                tryCount++;
            }
            while(true);
        }

        #endregion
        #region 메시지 요청하기 - RequestMessage(subject, message, timeout)

        /// <summary>
        /// 메시지 요청하기
        /// </summary>
        /// <param name="subject">수신 서브젝트</param>
        /// <param name="message">송신 메시지</param>
        /// <param name="timeout">타임아웃 (디폴트 : 30초)</param>
        /// <returns>응답 메시지</returns>
        public string RequestMessage(string subject, string message, int timeout = 30000)
        {
            byte[] messageByteArray = Encoding.UTF8.GetBytes(message);

            byte[] replyMessageByteArray = RequestMessage(subject, messageByteArray, timeout);

            string replyMessage = Encoding.UTF8.GetString(replyMessageByteArray);

            return replyMessage;
        }

        #endregion

        #region 리소스 해제하기 - Dispose()

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        public void Dispose()
        {
            Stop();
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 서버 발견시 이벤트 발생시키기 - FireServerDiscoveredEvent(connection)

        /// <summary>
        /// 서버 발견시 이벤트 발생시키기
        /// </summary>
        /// <param name="connection">NATS 연결</param>
        protected void FireServerDiscoveredEvent(Connection connection)
        {
            ConnectionInfo info = ConnectionInfo.GetConnectionInfo(connection);

            ServerDiscoveredEventArgs e = new ServerDiscoveredEventArgs(info);

            ServerDiscovered?.Invoke(this, e);
        }

        #endregion
        #region 서버 재연결시 이벤트 발생시키기 - FireServerReconnectedEvent(connection)

        /// <summary>
        /// 서버 재연결시 이벤트 발생시키기
        /// </summary>
        /// <param name="connection">NATS 연결</param>
        protected void FireServerReconnectedEvent(Connection connection)
        {
            ConnectionInfo info = ConnectionInfo.GetConnectionInfo(connection);

            ServerReconnectedEventArgs e = new ServerReconnectedEventArgs(info);

            ServerReconnected?.Invoke(this, e);
        }

        #endregion
        #region 메시지 수신시 이벤트 발생시키기 - FireMessageReceivedEvent(message)

        /// <summary>
        /// 메시지 수신시 이벤트 발생시키기
        /// </summary>
        /// <param name="message">NATS 메시지</param>
        protected void FireMessageReceivedEvent(Msg message)
        {
            MessageReceivedEventArgs e = new MessageReceivedEventArgs(message.Subject, message.Data, message.Reply);

            MessageReceived?.Invoke(this, e);

            if(e.ReplyMessageByteArray != null)
            {
                message.ArrivalSubcription.Connection.Publish(message.Reply, e.ReplyMessageByteArray);
            }
        }

        #endregion
        #region 서버 연결 중단시 이벤트 발생시키기 - FireServerDisconnectedEvent(connection)

        /// <summary>
        /// 서버 연결 중단시 이벤트 발생시키기
        /// </summary>
        /// <param name="connection">NATS 연결</param>
        protected void FireServerDisconnectedEvent(Connection connection)
        {
            ConnectionInfo info = ConnectionInfo.GetConnectionInfo(connection);

            ServerDisconnectedEventArgs e = new ServerDisconnectedEventArgs(info);

            ServerDisconnected?.Invoke(this, e);
        }

        #endregion
        #region 연결 폐쇄시 이벤트 발생시키기 - FireConnectionClosedEvent(connection)

        /// <summary>
        /// 연결 폐쇄시 이벤트 발생시키기
        /// </summary>
        /// <param name="connection">NATS 연결</param>
        protected void FireConnectionClosedEvent(Connection connection)
        {
            ConnectionInfo info = ConnectionInfo.GetConnectionInfo(connection);

            ConnectionClosedEventArgs e = new ConnectionClosedEventArgs(info);

            ConnectionClosed?.Invoke(this, e);
        }

        #endregion
        #region 에러 이벤트 발생시키기 - FireErrorEvent(e)

        /// <summary>
        /// 에러 이벤트 발생시키기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected void FireErrorEvent(ErrEventArgs e)
        {
            ConnectionInfo connectionInfo = ConnectionInfo.GetConnectionInfo(e.Conn);

            string key = SubscriptionInfo.GetKey(e.Subscription.Subject, e.Subscription.Queue);

            SubscriptionInfo subscriptionInfo = null;

            if(this.subscriptionInfoDictionary.ContainsKey(key))
            {
                subscriptionInfo = this.subscriptionInfoDictionary[key];
            }

            string error = e.Error;

            ErrorEventArgs errorEventArgs = new ErrorEventArgs(connectionInfo, subscriptionInfo, error);

            Error?.Invoke(this, errorEventArgs);
        }

        #endregion
    }
}