﻿using System;

using NATS.Client;

namespace TestCommon
{
    /// <summary>
    /// 서버 정보
    /// </summary>
    public class ServerInfo : IDisposable
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// URL
        /// </summary>
        private string url;

        /// <summary>
        /// 사용자 ID
        /// </summary>
        private string userID;

        /// <summary>
        /// 패스워드
        /// </summary>
        private string password;

        /// <summary>
        /// PINT 주기
        /// </summary>
        private int pingInterval;

        /// <summary>
        /// 클러스터 URL 배열
        /// </summary>
        private string[] clusterURLArray;

        /// <summary>
        /// NATS 연결
        /// </summary>
        private IConnection connection;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region URL - URL

        /// <summary>
        /// URL
        /// </summary>
        public string URL
        {
            get
            {
                return this.url;
            }
        }

        #endregion
        #region 사용자 ID - UserID

        /// <summary>
        /// 사용자 ID
        /// </summary>
        public string UserID
        {
            get
            {
                return this.userID;
            }
        }

        #endregion
        #region 패스워드 - Password

        /// <summary>
        /// 패스워드
        /// </summary>
        public string Password
        {
            get
            {
                return this.password;
            }
        }

        #endregion
        #region PING 주기 - PingInterval

        /// <summary>
        /// PING 주기
        /// </summary>
        public int PingInterval
        {
            get
            {
                return this.pingInterval;
            }
        }

        #endregion
        #region 클러스터 URL 배열 - ClusterURLArray

        /// <summary>
        /// 클러스터 URL 배열
        /// </summary>
        public string[] ClusterURLArray
        {
            get
            {
                return this.clusterURLArray;
            }
        }

        #endregion
        #region NATS 연결 - Connection

        /// <summary>
        /// NATS 연결
        /// </summary>
        public IConnection Connection
        {
            get
            {
                return this.connection;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - ServerInfo(url, userID, password, pingInterval, clusterURLArray)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="url">URL</param>
        /// <param name="userID">사용자 ID</param>
        /// <param name="password">패스워드</param>
        /// <param name="pingInterval">PING 주기</param>
        /// <param name="clusterURLArray">클러스터 URL 배열</param>
        public ServerInfo(string url, string userID, string password, int pingInterval, string[] clusterURLArray = null)
        {
            if(string.IsNullOrWhiteSpace(url))
            {
                throw new ArgumentException("url이 null이거나 공백 문자열 입니다.");
            }

            if(pingInterval < 100)
            {
                throw new ArgumentException("pingInterval은 100 이상이어야 합니다.");
            }

            this.url = url.Trim();

            this.userID   = string.IsNullOrWhiteSpace(userID  ) ? null : userID.Trim();
            this.password = string.IsNullOrWhiteSpace(password) ? null : password.Trim();

            this.pingInterval = pingInterval;

            if(clusterURLArray == null)
            {
                this.clusterURLArray = null;
            }
            else
            {
                if(clusterURLArray.Length > 0)
                {
                    this.clusterURLArray = clusterURLArray;
                }
                else
                {
                    this.clusterURLArray = null;
                }
            }

            this.connection = null;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 생성자 - ServerInfo()

        /// <summary>
        /// 생성자
        /// </summary>
        private ServerInfo()
        {
        }

        #endregion

        #region NATS 연결 시작하기 - StartConnection(options)

        /// <summary>
        /// NATS 연결 시작하기
        /// </summary>
        /// <param name="options">옵션</param>
        public void StartConnection(Options options)
        {
            this.connection = new ConnectionFactory().CreateConnection(options);
        }

        #endregion
        #region NATS 연결 중단하기 - StopConnection()

        /// <summary>
        /// NATS 연결 중단하기
        /// </summary>
        public void StopConnection()
        {
            if(this.connection != null)
            {
                this.connection.Dispose();
            }
        }

        #endregion

        #region NATS 옵션 구하기 - GetOptions()

        /// <summary>
        /// NATS 옵션 구하기
        /// </summary>
        /// <returns>NATS 옵션</returns>
        public Options GetOptions()
        {
            Options options = ConnectionFactory.GetDefaultOptions();

            options.Url          = this.url;
            options.User         = this.userID;
            options.Password     = this.password;
            options.Servers      = this.clusterURLArray;
            options.PingInterval = this.pingInterval;

            return options;
        }

        #endregion

        #region 리소스 해제하기 - Dispose()

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        public void Dispose()
        {
            StopConnection();
        }

        #endregion
    }
}